/* Hobbit's efforts at a unified s/key. */

/* Snarfed from Bellcore/NRL s/key distribution, berkeley "su" 5.4, and
   unified.  So named because it can be skeysh, or skeysu, or skeyin --
   either way, it's a barrier layer, and if you don't penetrate it just
   right, it bleeds all over the syslog.

   This combines the sources of "skeysh", "su", and a new twist, "skeyin".
   This is because the functions of these are so similar and contain a lot
   of common code; it's easier to #ifdef the right things.

   Compile with one of SU, SH, or IN defined; if IN, you also need to
   specify SKEYUID as the UID of your captive "skey" account.  Also, check
   _PATH_various below if your system is strange.

   If USE_OWN is defined, "su" uses the caller's skey for becoming root
   instead of root's skey.  */

#include "generic.h"

#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef UTMPX		/* xxx: does utmpx have everything utmp needs? */
#include <utmpx.h>
struct utmpx utmp;
struct utmpx *utxp;
#else
#include <utmp.h>
struct utmp utmp;
#endif /* UTMPX */

#ifdef HAVE_PATHS_H
#include <paths.h>
#endif
/* xxx: Some of these might be in /etc or /var/run or /var/log or something */
#ifndef _PATH_LOGIN
#define _PATH_LOGIN "/bin/login"
#endif
#ifndef _PATH_UTMP
#define _PATH_UTMP  "/etc/utmp"
#endif
#ifndef _PATH_WTMP
#define _PATH_WTMP  "/usr/adm/wtmp"
#endif

#ifdef HAVE_STRINGS_H	/* what a swamp, sometimes */
#include <strings.h>
/* uncomment if your .h files are fucked up:
#define index	strchr
#define rindex	strrchr
#define bzero(s,l) memset(s,0,l)
*/
#else
#include <string.h>
#endif /* STRINGS_H */

#define	RETRIES 2	/* skeyin login attempts */

char	userbuf[16]	= "USER=";
char	homebuf[128]	= "HOME=";
char	shellbuf[128]	= "SHELL=";
/* Updated all of this to be much more paranoid.  Who the HELL had
   dot first in the PATH, fer instance?!  _H*/
char	pathbuf[128]	= "PATH=/bin:/usr/bin:/usr/ucb:/sbin:/usr/sbin";
char	*cleanenv[] = { userbuf, homebuf, shellbuf, pathbuf, 0, 0 };
#ifdef SU
char	User[16] = "root";	/* xxx: how big are these, really?! */
#else
char	User[16] = "nobody";	/* Answer: exactly as you CREATE them. */
#endif
char	*user = User;		/* and *then* point to it.  BSD bozoids. */
char	Shell[128] = "/invalid!";
char	*shell = Shell;
char	host[256] = "LOCAL";	/* big, so utmpx stuff works */
char	*who;			/* xtra */
int	fulllogin;
int	fastlogin;

extern char	**environ;
struct	passwd *pwd;
char	*crypt();
char	*getpass();
char	*getenv();
char	*getlogin();
char	*ttyname();

#ifndef HAVE_SETPRIORITY
#define setpriority(x,y,z)      ;
#endif

/* figure out what we're being compiled as */
#undef PROG
#ifdef SU
#define PROG "skeysu"
#endif
#ifdef SH
#define PROG "skeysh"
#endif
#ifdef IN
#define PROG "skeyin"
#endif
#ifndef PROG
BAILING: One of SU, SH, or IN must be defined!
#endif

int ufd;
int ufoff = 0;
char prog[16];
void rip();		/* in skeysubr.c */

main(argc,argv)
	int argc;
	char *argv[];
{
	char *password;
	struct group *gr;
	char buf[1000];
	FILE *fp;
	register char *p;
        int stat;
	int tries;
	int i;
	
#ifdef LOG_AUTH
	openlog(PROG, LOG_ODELAY, LOG_AUTH);
#else /* probably buglix */
	openlog (PROG, 0);
#endif /* syslog type */
	strcpy (prog, PROG);

#ifdef IN
#ifndef SKEYUID
BAILING AGAIN: SKEYUID must be defined for "skeyin"!
#endif

/* skeyin does several checks before anything else.  For example, rlogin
   and rexec hand in arguments, even if they're null; skeyin
   obviously won't work there... */
	if (argc > 1) {
		fprintf (stderr, "%s: Invalid argument\n", prog);
		bail(1);
	}
#ifndef TEST
	if (getuid() != SKEYUID) {
		fprintf (stderr, "%s: Can't run as this UID\n", prog);
		syslog (LOG_WARNING, "UID %d tried to run", getuid());
		bail(1);
	}
#endif /* TEST */
#endif /* IN */
#ifndef TEST
	if (geteuid() != 0) {				/* ALL ... */
		fprintf (stderr, "%s: Must be setuid root\n", prog);
		bail(1);
	}
#endif /* TEST */
	if ((pwd = getpwuid(getuid())) == NULL) {
		fprintf(stderr, "Who are you?\n");
		syslog (LOG_WARNING, "Unknown UID %d tried to run", getuid());
		bail(1);
	}
	strcpy(buf, pwd->pw_name);		/* buf: calling user */
	getutmph (host);

/* Fill in "user" one of three ways: su-arg, sh-caller, or ask for it */

#ifdef SU
again:
	if (argc > 1 && strcmp(argv[1], "-f") == 0) {
		fastlogin++;
		argc--, argv++;
		goto again;
	}
	if (argc > 1 && strcmp(argv[1], "-") == 0) {
		fulllogin++;
		argc--, argv++;
		goto again;
	}
	if (argc > 1 && argv[1][0] != '-') {
		user = argv[1];			/* user: target user */
		argc--, argv++;
	}
#endif /* SU */

#ifdef SH
	strcpy (user, pwd->pw_name);		/* user: caller AND target */
#endif

#ifdef IN
	tries = 1;
uagain:
	if (feof (stdin))
	  bail (0);
	if (tries > RETRIES)
	  bail (-1);
	printf ("User: ");
	fflush (stdout);
	if ((fgets (user, 10, stdin)) == NULL)
	  goto uagain;
	rip (user);
	if (*user == '\0')
	  goto uagain;
	tries++;
/* xxx: do we need a timeout in here?  A LONG one, if at all... */

#endif /* IN */

/* Done with calling user details; safe to overwrite with target pwd now */
	if ((pwd = getpwnam(user)) == NULL) {
		fprintf(stderr, "Unknown login: %s\n", user);
#ifndef IN
		bail(1);
#else
		goto uagain;
/* XXX: I *really* wanna do something like this instead.
		skey_fakeout (user);
		syslog (LOG_WARNING, "Fakeout: %s on %s from %s",
		  user, ttyname(2), host);
		goto uagain;
/* Also, if they try to login as SKEYUID again, get abusive. */
#endif /* not IN */
	} /* if new pwd (user); we should now have a valid one */
	who = user;

#ifndef SH
/* if target user is root or equivalent, do some checks */
	if (pwd->pw_uid == 0) {			/* NEW pwd contents */

#ifdef IN				/* Read my lips: No root logins! */
/* XXX: better would be to check uid < some-magic-cutoff */
		fprintf (stderr, "Permission denied\n");
		syslog (LOG_WARNING, "ROOT LOGIN ATTEMPT on %s from %s",
		  ttyname(2), host);
		bail (-1);
#endif /* IN */
#ifdef SU
#ifndef TEST
		if ((gr = getgrgid(0)) != NULL) {
		    for (i = 0; gr->gr_mem[i] != NULL; i++)
			  if (strcmp(buf, gr->gr_mem[i]) == 0)
			    goto userok;
		    fprintf(stderr, "You do not have permission to su %s\n",
				user);
		    syslog (LOG_WARNING, "SU ATTEMPT by %s on %s from %s",
			buf, ttyname(2), host);
		    bail (1);
		}
	userok:
#endif /* TEST */
#ifdef USE_OWN
	/* optional: use *calling* user's skey for root "su" */
	who = buf;
#endif /* USE_OWN */
#endif /* SU */
		setpriority(PRIO_PROCESS, 0, -2);
	} /* if target uid = 0 */
#endif /* not SH */

#define Getlogin()  (((p = getlogin()) && *p) ? p : buf)

        stat = skey_haskey (who);		/* ALL */

        if (stat == 1) {
	  fprintf(stderr,"%s: No s/key entry for user %s.\n", prog, who);
	  bail(1);
/*           goto OLDPASSWDENTRY; */
	}

        if (stat == -1) {
	  fprintf(stderr, "%s: Could not open key file.\n", prog);
	  bail (-1);
/*           goto OLDPASSWDENTRY; */
	}

        if (skey_authenticate (who) == -1) {

#ifdef SU
	  fprintf(stderr, "Sorry\n");
	  if (pwd->pw_uid == 0) {
                    syslog (LOG_WARNING, "BAD SU %s on %s from %s",
                          Getlogin(), ttyname(2), host);
	  }
#endif
#ifdef IN
	  fprintf (stderr, "login incorrect\n");
	  syslog (LOG_WARNING, "BAD LOGIN %s on %s from %s",
	      user, ttyname(2), host);
	  goto uagain;
#else
	  bail (-1);
#endif /* IN */
        } /* if not skey_auth */

#if 0
        goto ok;

/* OLDPASSWDENTRY:			/* dont allow any of this */

	if (pwd->pw_passwd[0] == '\0' || getuid() == 0)
		goto ok;
	password = getpass("Password:");
	if (strcmp(pwd->pw_passwd, crypt(password, pwd->pw_passwd)) != 0) {
		fprintf(stderr, "Sorry\n");
		if (pwd->pw_uid == 0) {
			syslog(LOG_WARNING, "BAD SU %s on %s",
					Getlogin(), ttyname(2));
		}
		exit(2);
	}
ok:
#endif /* 0, disabled */

	endpwent();
	umask (0022);

/* Cleared. You may proceed... */

#ifdef IN
	syslog (LOG_NOTICE, "%s on %s from %s", user, ttyname(2), host);
	closelog();
	fulllogin = 1;
/* XXX: there might be a variety of grungy hacks here to get login to
   believe us.  Use "dumpinvoc" to figure out which one it is.  If there
   isn't one, you have to turn on LOGIN_HAIR and pray.  _H*/
#ifdef LOGIN_HAIR
	login_hair (user);	/* fix up utmp/wtmp/lastlog/tty/the world */
#else
	setuid (0);		/* have to really call login as root */
/* xxx: this is UNTESTED.  Feedback encouraged.  _H*/
	execl (_PATH_LOGIN, "login", "-f", user,
		"-h", host, (char *) 0);
	fprintf (stderr, "%s: login failed\n", prog);
/* this is noteworthy; log it even tho we closed before */
	syslog (LOG_WARNING, "real login exec FAILED for %s", user);
	bail (-1);
#endif /* LOGIN_HAIR */

#else /* IN */
/* XXX: should syslog anyways, at least for SU */
	if (pwd->pw_uid == 0) {
		syslog(LOG_NOTICE, "%s on %s from %s", Getlogin(),
		  ttyname(2), host);
	}
#endif /* IN */

/* ALL fall here, including IN if we have to finish up the login hair */
#ifndef TEST
	if (setgid(pwd->pw_gid) < 0) {
		perror("setgid");
		bail(3);
	}
	if (initgroups(user, pwd->pw_gid)) {
		fprintf(stderr, "%s: initgroups failed\n", prog);
		bail(4);
	}
	if (setuid(pwd->pw_uid) < 0) {
		perror("setuid");
		bail(5);
	}
/* This was a HOLE if shell is initialized as "/bin/sh".  Fixed.  _H*/
	if (pwd->pw_shell && *pwd->pw_shell)
		shell = pwd->pw_shell;
	else {
		fprintf (stderr, "No shell\n");
		bail(7);
	}
#endif /* not TEST */
#ifdef SH
	fulllogin = 1;
	fastlogin = 0;
/* XXX, sorta; we might want "/foo/bar/skXXX" to become "/bin/XXX" as
   discussed at least once on skey-users */
	strcpy (shell, "/bin/csh");
#endif
	if (fulllogin) {
		cleanenv[4] = getenv("TERM");
		environ = cleanenv;
	}
	if (strcmp(user, "root"))
		setenv("USER", pwd->pw_name, userbuf);
	setenv("SHELL", shell, shellbuf);
	setenv("HOME", pwd->pw_dir, homebuf);
	setpriority(PRIO_PROCESS, 0, 0);
	closelog();

	if (fastlogin) {
		*argv-- = "-f";
		*argv = "su";
	} else if (fulllogin) {
		if (chdir(pwd->pw_dir) < 0) {
			fprintf(stderr, "No directory\n");
			if (chdir ("/"))
				bail(6);
		}
#ifdef SU
		*argv = "-su";
#else
		*argv = "-csh";
#endif
	} else /* fulllogin */
		*argv = "su";
	execv(shell, argv);
	fprintf(stderr, "No shell\n");
	bail(7);
} /* main */

#ifdef XTRA	/* maybe wanna do something with this code; forms "-csh"? */
/* from bellcore skey login.c */
/* xxx: use strrchr */
	if ((namep = rindex(pwd->pw_shell, '/')) == NULL)
		namep = pwd->pw_shell;
	else
		namep++;
	strcat(minusnam, namep);	/* and exec with that */
#endif /* XTRA */

setenv(ename, eval, buf)
	char *ename, *eval, *buf;
{
	register char *cp, *dp;
	register char **ep = environ;

	/*
	 * this assumes an environment variable "ename" already exists
	 */
	while (dp = *ep++) {
		for (cp = ename; *cp == *dp && *cp; cp++, dp++)
			continue;
		if (*cp == 0 && (*dp == '=' || *dp == 0)) {
			strcat(buf, eval);
			*--ep = buf;
			return;
		}
	}
}

char *
getenv(ename)
	char *ename;
{
	register char *cp, *dp;
	register char **ep = environ;

	while (dp = *ep++) {
		for (cp = ename; *cp == *dp && *cp; cp++, dp++)
			continue;
		if (*cp == 0 && (*dp == '=' || *dp == 0))
			return (*--ep);
	}
	return ((char *)0);
}

bail (stat)
int stat;
{
/* wrap up everything, make sure it gets out, and get out ourselves */
  fflush (stdout);
  fflush (stderr);
  endpwent();
  closelog();
  sleep (1);
  exit (stat);
}

#ifdef LOGIN_HAIR

/* If -f doesn't work, we have to go through the motions of login ourselves.
   Much of this swiped out of an SV-compat login in some other s/key dist.
   If you're very lucky, the stock login on a target box you come up against
   *will* handle -f and you won't have to worry about this nonsense. */

#define	SCPYN(a, b)	strncpy(a, b, sizeof(a))
#define TTYGRPNAME	"tty"		/* XXX */
#define sleepexit	bail

login_hair(username)
char *username;
{
struct utmp clrutmp;

/* some systems dont have lastlog at all, so be sure to #undef _PATH_LASTLOG
   on such machines.  Lastlog code is switched on this... */
/* xxx: "generic.h" and /usr/adm/lastlog or /var/log/lastlog ... */
/* DISABLED:
#ifndef _PATH_LASTLOG
#define _PATH_LASTLOG "/usr/adm/lastlog"
#endif
*/
#ifdef HAVE_LASTLOG_H
#include <lastlog.h>	/* All this might be in utmp.h instead */
#endif
#ifdef AIX		/* Feh!  Twits. */
#include <login.h>
#endif
#ifdef _PATH_LASTLOG
struct lastlog ll;	/* xxx: where do we get this from? */
#endif

struct group *gr;
char *ttyn, *tty;
int f;
int fd;
int t;

/* We hit this already logged in as "skey" or whatever with a corresponding
   utmp/utmpx entry, and only need to fix:
	utmp: ut_name, ut_time;  host and line are already set
	wtmp: close entry for "skey"; open new one for user
	lastlog: the usual stuff
   The environment will be squashed later.
*/
	ttyn = ttyname(0);
	if (ttyn == NULL || *ttyn == '\0')
		ttyn = "/dev/tty??";
/* appropriately strip either /dev/ttypx and /dev/pts/x */
/* XXX: probably won't work on hpux "ptmx" stuff. */
	if (tty = strchr(ttyn + 1, '/'))
		++tty;
	else
		tty = ttyn;

/* largely swiped Wietse's sysv stuff for this, I just don't wanna deal */
#ifdef UTMPX
/* Update SYSV-style utmp and wtmp files. */
	if (utmpx_swap (tty, user) != 0) {
		fprintf (stderr, "%s: No utmpx entry\n");
		syslog (LOG_WARNING, "No utmpx entry for %s on %s",
		  user, tty);
		sleepexit(1);
	}
#else /* UTMPX */

/* Update BSD-style utmp and wtmp files. */
/* do NOT zero the utmp first; we're only making changes */
	(void)time(&utmp.ut_time);
	strncpy(utmp.ut_name, user, sizeof(utmp.ut_name));
	if (host)
		strncpy(utmp.ut_host, host, sizeof(utmp.ut_host));
	strncpy(utmp.ut_line, tty, sizeof(utmp.ut_line));

/* copy completed utmp to an expendable one, and clear user and host */
	memcpy ((char *)&clrutmp, (char *)&utmp, sizeof (utmp));
	memset (clrutmp.ut_name, 0, sizeof (utmp.ut_name));
	memset (clrutmp.ut_host, 0, sizeof (utmp.ut_host));

/* and that same ol' tired login code from 1987 ... */
	t = ttyslot();
	if (t > 0 && (f = open(_PATH_UTMP, O_WRONLY)) >= 0) {
		lseek(f, (long)(t*sizeof(utmp)), 0);
		SCPYN(utmp.ut_line, tty);
		write(f, (char *)&utmp, sizeof(utmp));
		close(f);
	}
	if ((f = open(_PATH_WTMP, O_WRONLY|O_APPEND)) >= 0) {
/* "log out" the skey pseudouser first */
		write(f, (char *)&clrutmp, sizeof (utmp));
		write(f, (char *)&utmp, sizeof(utmp));
		close(f);
	}
/* It is now clear to me that the lack of locking, and the lack of concern
   over having wtmp-updating be one clean atomic operation, is often
   responsible for interesting output from "last".  We tried to make it
   quick here... */

#endif /* UTMPX */

#ifdef _PATH_LASTLOG
/* lastlog stuff for target user.  Ignore .hushlogin for this */
	if ((fd = open(_PATH_LASTLOG, O_RDWR, 0)) >= 0) {
		(void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), 0);
		if (read(fd, (char *)&ll, sizeof(ll)) == sizeof(ll) &&
		    ll.ll_time != 0) {
#if 0
#ifdef __svr4__			/* Last login too long ago? */
			if (pwd->pw_uid && spwd->sp_inact > 0 
			    && ll.ll_time / DAY
			    + spwd->sp_inact < DAY_NOW) {
				printf("Your account has been inactive to long.\n");
				sleepexit(1);
			}
#endif /* __svr4__ */
#endif /* disabled... */
			printf("Last login: %.*s ",
			    24-5, (char *)ctime(&ll.ll_time));
			if (*ll.ll_host != '\0')
				printf("from %.*s\n",
				    sizeof(ll.ll_host), ll.ll_host);
			else
				printf("on %.*s\n",
				    sizeof(ll.ll_line), ll.ll_line);
		} /* read && time */
		(void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), 0);
		bzero((char *)&ll, sizeof(ll));
		(void)time(&ll.ll_time);
		strncpy(ll.ll_line, tty, sizeof(ll.ll_line));
		if (host)
			/* strncpy(ll.ll_host, host, 16); */
			/* xxx: how big *is* it?? */
			strncpy(ll.ll_host, host, sizeof (ll.ll_host));
	       (void)write(fd, (char *)&ll, sizeof(ll));
	       (void)close(fd);
	} /* open lastlog */
#endif /* _PATH_LASTLOG */

/* xxx: may error out if "/dev/tty??" !! */
	(void)chown(ttyn, pwd->pw_uid,
	    (gr = getgrnam(TTYGRPNAME)) ? gr->gr_gid : pwd->pw_gid);
	(void)chmod(ttyn, 0620);

} /* login_hair ...  Everything else is handled by caller. */
#endif /* LOGIN_HAIR */

/* Pull our hostname out of utmp -- in all cases.  Potentially useful, and
   utmp-grubbing is the best we can do, considering we're usually on the
   wrong end of a pty and can't getpeername. */

getutmph (buf)
char *buf;
{	int stat;
	int x;
	char devent[16];
	char *dp;

/*	x = ttyslot();		/* xxx, kinda; do we trust this? ans: NO. */
#ifdef DEBUG
printf ("ttyslot seek: %d\n", x);
#endif
/*	lseek (ufd, (long) x * sizeof (utmp), 0); */

/* do it the hard way, which ttyslot() does anyways if it works at all */
	strcpy (devent, ttyname(0));
	dp = devent + 4;
#ifdef DEBUG
printf ("ttyname str: *%s* ... ", devent);
#endif
	*dp = '\0';
	if (strcmp (devent, "/dev")) {
		fprintf (stderr, "%s: ttyname screwed up: %s\n", prog,
		  devent);
	}
	dp++;
	x = strlen (dp);

#ifdef UTMPX
	strncpy (utmp.ut_line, dp, sizeof (utmp.ut_line));
	if ((utxp = getutxline (&utmp)) == 0) {
		fprintf (stderr, "%s: can't find utmpx entry\n", prog);
		syslog (LOG_WARNING, "No utmpx entry for %s", dp);
		bail (-1);	/* xxx: is that really a hairy error? */
	}
	if (*utxp->ut_host) {
		memset (host, 0, 256);
#ifdef DEBUG
printf ("utxp %x utxp->host %x : %s\n", utxp, utxp->ut_host, utxp->ut_host);
#endif
		strncpy (host, utxp->ut_host, utxp->ut_syslen);
	}
	endutxent();
#else /* UTMPX */
	ufd = (open (_PATH_UTMP, O_RDONLY));
	if (ufd == -1) {
		fprintf (stderr, "%s: can't open utmp\n", prog);
		syslog (LOG_WARNING, "can't open utmp");
		bail (-1);	/* xxx: hairy error? */
	}
	lseek (ufd, 0, 0);
	ufoff = 0;
	while ((stat = read (ufd, &utmp, sizeof (utmp))) > 0) {
#ifdef DEBUG
printf ("ut: %s x is %d", utmp.ut_line, x);
#endif
		if (strncmp (utmp.ut_line, dp, x) == 0)
		  if (utmp.ut_host[0]) {
		    memset (host, 0, 20);
		    strncpy (host, utmp.ut_host, sizeof (utmp.ut_host));
		    break;
		  }
		ufoff += sizeof (utmp);	/* save for later */
	} /* while read */
	close (ufd);
#endif /* UTMPX */

#ifdef DEBUG
printf ("got it: %s ufoff %d\n", utmp.ut_line, ufoff);
printf ("Whole utmp: ");  write (0, &utmp, sizeof (utmp));
printf ("\nHost: %s\n", host); fflush (stdout);
#endif

} /* getutmph */
